/* gTask
 * Copyright (C) 2004 Michael Henson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-connection-group.h>

#include <string.h>

/** @addtogroup GTaskConnectionGroup */

/* @{ */

static gpointer parent_class = NULL;

static void
gtask_connection_group_class_init( GTaskConnectionGroupClass *klass );

static void
gtask_connection_group_init( GTaskConnectionGroup *group, gpointer g_class );

static void
gtask_connection_group_dispose( GObject *obj );

static void
gtask_connection_group_finalize( GObject *obj );

/**
 * @brief registers the GTaskConnectionGroup object with the GObject system
 *        if that has yet to be done and then returns the GType for
 *        GTaskConnectionGroup
 *
 * @author Michael Henson
 *
 * @return the GType object for GTaskConnectionGroup
 */
GType gtask_connection_group_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskConnectionGroupClass ),
            NULL,   /* base init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_connection_group_class_init, /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskConnectionGroup ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_connection_group_init
        };

        type = g_type_register_static( G_TYPE_OBJECT,
                                       "GTaskConnectionGroup",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_connection_group_class_init( GTaskConnectionGroupClass *klass ) {
    GObjectClass *gobject_class = G_OBJECT_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->dispose = gtask_connection_group_dispose;
    gobject_class->finalize = gtask_connection_group_finalize;
}

/**
 * @brief initializes a GTaskConnectionGroup
 *
 * @author Michael Henson
 */
static void
gtask_connection_group_init( GTaskConnectionGroup *group, gpointer g_class ) {
    GTaskConnectionGroupPrivate *private;

    private = g_new0( GTaskConnectionGroupPrivate, 1 );

    private->connection_list = NULL;
    private->dispose_has_run = FALSE;

    group->private = private;
}

static void
gtask_connection_group_dispose( GObject *obj ) {
    GTaskConnectionGroup *group = GTASK_CONNECTION_GROUP( obj );

    if( group->private->dispose_has_run )
        return;

    g_slist_foreach( group->private->connection_list,
                     (GFunc) g_object_unref,
                     NULL );

    group->private->dispose_has_run = TRUE;

    if( G_OBJECT_CLASS( parent_class )->dispose )
        G_OBJECT_CLASS( parent_class )->dispose( obj );
}

static void
gtask_connection_group_finalize( GObject *obj ) {
    GTaskConnectionGroup *group = GTASK_CONNECTION_GROUP( obj );

    g_slist_free( group->private->connection_list );

    g_free( group->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

/**
 * @brief creates a new GTaskConnectionGroup
 *
 * @author Michael Henson
 *
 * @return the new GTaskConnectionGroup
 */
GTaskConnectionGroup *
gtask_connection_group_new( void ) {
    return g_object_new( GTASK_CONNECTION_GROUP_TYPE, NULL );
}

/**
 * @brief Adds in the passed in GTaskConnection to the GTaskConnectionGroup.
 *
 * @attention The passed in GTaskConnection will have its reference count
 *              increased by one.
 *
 * @author Michael Henson
 *
 * @param group a GTaskConnectionGroup
 * @param conn a GTaskConnection to add to the group
 *
 * @todo should check and see if the passed in connection
 *       is already in the group
 */
void
gtask_connection_group_add_connection( GTaskConnectionGroup *group,
                                       GTaskConnection *conn )
{
    GTaskConnectionGroupPrivate *private;

    g_return_if_fail( GTASK_IS_CONNECTION_GROUP( group ) );
    g_return_if_fail( GTASK_IS_CONNECTION( conn ) );
    g_return_if_fail( !group->private->dispose_has_run );

    g_object_ref( conn );

    private = group->private;

    private->connection_list = g_slist_append( private->connection_list, conn );
}

/**
 * @brief Removes all GTaskConnections from the passed in GTaskConnectionGroup.
 *
 * @author Michael Henson
 *
 * @param group a GTaskConnectionGroup
 */
void
gtask_connection_group_clear( GTaskConnectionGroup *group ) {
    GTaskConnectionGroupPrivate *private;

    g_return_if_fail( GTASK_IS_CONNECTION_GROUP( group ) );
    g_return_if_fail( !group->private->dispose_has_run );

    private = group->private;

    g_slist_foreach( private->connection_list,
                     (GFunc) g_object_unref,
                     NULL );

    g_slist_free( private->connection_list );

    private->connection_list = NULL;
}

/**
 * @brief Returns the internal GSList that holds all of the GTaskConnections.
 *
 * @author Michael Henson
 *
 * @param group a GTaskConnectionGroup
 *
 * @return a GSList if there are connections in the group, NULL otherwise.
 *         The returned list should be considered read only in nature.
 */
GSList *
gtask_connection_group_get_list( GTaskConnectionGroup *group ) {
    g_return_val_if_fail( GTASK_IS_CONNECTION_GROUP( group ), NULL );
    g_return_val_if_fail( !group->private->dispose_has_run, NULL );

    return group->private->connection_list;
}

/** @brief Searches through the GTaskConnectionGroup's internal list and
 *         returns the GTaskConnection whose name matches the passed in name.
 *
 * @param group A GTaskConnectionGroup
 * @param conn_name the name of the connection to return
 *
 * @return the GTaskConnection or NULL of no such connection with the
 *         passed in name was found.
 */
GTaskConnection *
gtask_connection_group_get_connection_by_name( GTaskConnectionGroup *group,
                                               const gchar *conn_name )
{
    GSList          *list;
    GTaskConnection *conn = NULL;

    g_return_val_if_fail( GTASK_IS_CONNECTION_GROUP( group ), NULL );
    g_return_val_if_fail( conn_name != NULL, NULL );
    g_return_val_if_fail( !group->private->dispose_has_run, NULL );

    list = group->private->connection_list;

    while( list ) {
        GTaskConnection *t_conn = (GTaskConnection *) list->data;
        const gchar     *name = gtask_connection_get_name( t_conn );

        if( strcmp( name, conn_name ) == 0 ) {
            conn = t_conn;
            break;
        }

        list = g_slist_next( list );
    }

    return conn;
}

/* @} */
