/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef GTASK_CONNECTION_H
#define GTASK_CONNECTION_H

#include <stdio.h>
#include <glib-2.0/glib-object.h>
#include <libxml/xmlIO.h>

#include <gtask/gtask-stream-parser.h>
#include <gtask/gtask-event-dispatcher.h>

/** @addtogroup GTaskConnection */

/* @{ */

typedef struct _GTaskConnection        GTaskConnection;
typedef struct _GTaskConnectionPrivate GTaskConnectionPrivate;
typedef struct _GTaskConnectionClass   GTaskConnectionClass;

/** returns the GType of GTaskConnection */
#define GTASK_CONNECTION_TYPE            (gtask_connection_get_type())

/** casts the passed in object to a GTaskConnection */
#define GTASK_CONNECTION(obj)            (G_TYPE_CHECK_INSTANCE_CAST( (obj), \
                                          GTASK_CONNECTION_TYPE, \
                                          GTaskConnection ))

/** casts the passed in object to a GTaskConnectionClass */
#define GTASK_CONNECTION_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST( (klass), \
                                          GTASK_CONNECTION_TYPE, \
                                          GTaskConnectionClass ))

/**
 * returns TRUE if the passed in object is a GTaskConnection,
 * FALSE otherwise
 */
#define GTASK_IS_CONNECTION(obj)         (G_TYPE_CHECK_INSTANCE_TYPE( (obj), \
                                          GTASK_CONNECTION_TYPE ))

/**
 * returns TRUE if the passed in object is a GTaskConnectionClass,
 * FALSE otherwise
 */
#define GTASK_IS_CONNECTION_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE( (klass), \
                                          GTASK_CONNECTION_TYPE ))

#define GTASK_CONNECTION_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS( (obj), \
                                          GTASK_CONNECTION_TYPE, \
                                          GTaskConnectionClass ))

typedef enum  {
    GTASK_ROLE_NONE,        ///< no role has been specified
    GTASK_ROLE_PROXY,       ///< the proxy role -- callbacks, etc are sent back to another gtask server for exectution
    GTASK_ROLE_REPORTER,    ///< the reporter role -- the ui displaying tasks
    GTASK_ROLE_PROVIDER     ///< the provider role -- the program creating tasks
} GTaskRole;

typedef enum {
    GTASK_CONNECTION_METHOD_SOCKET,
    GTASK_CONNECTION_METHOD_LOOPBACK,
    GTASK_CONNECTION_METHOD_FD,
    GTASK_CONNECTION_METHOD_UNKNOWN
} GTaskConnectionMethod;

#define GTASK_ROLE_TYPE                   (gtask_role_get_type())

/**
 * A struct to hold private information for a GTaskConnection.
 */
struct _GTaskConnectionPrivate {
    gint                    inp_fd;     ///< a unix file descriptor (input)
    gint                    out_fd;     ///< a unix file descriptor (output)

    xmlOutputBufferPtr      xml_output; ///< xml output buffer

    GIOChannel              *g_input;

    GTaskRole                role;
    GString                 *name;

    gboolean                 g_hup_watch_on;
    guint                    g_hup_source_id;
    guint                    g_in_source_id;

    GTaskStreamParser       *parser;
    GTaskEventDispatcher    *dispatcher;

    /* how did we get connected in the first place? */
    GTaskConnectionMethod    conn_method;
    char                    *socket_uri;

    gboolean                 auto_reconnect;
    gboolean                 auto_parse;
};

/**
 * A struct intended to represent a connection to a gtask daemon.
 * It holds basic communication information and is used by 
 * GTaskGenericTask, etc to make notifications to the server.
 */
struct _GTaskConnection {
    GObject                 parent;     ///< the connection's parent
    GTaskConnectionPrivate *private;    ///< the connection's private data
};

struct _GTaskConnectionClass {
    GObjectClass parent;

    guint        connected_signal_id;
    guint        disconnected_signal_id;
};

GType gtask_role_get_type( void );

GType gtask_connection_get_type( void );

char *
gtask_connection_default_server_bind_path( );

GTaskConnection *
gtask_connection_new( void );

GTaskConnection *
gtask_connection_new_with_properties( const gchar *name, GTaskRole role );

char *
gtask_connection_default_server_bind_path( void );

gboolean
gtask_connection_connect_domain_socket_default( GTaskConnection *conn );

gboolean
gtask_connection_connect_domain_socket( GTaskConnection *conn, const char *file );

gboolean
gtask_connection_connect_fd( GTaskConnection *conn, int inp_fd, int out_fd );

gboolean
gtask_connection_connect_loopback( GTaskConnection *conn );

gboolean
gtask_connection_is_connection( GTaskConnection *conn );

gboolean gtask_connection_close( GTaskConnection *conn );

int
gtask_connection_get_input_fd( GTaskConnection *conn );

int
gtask_connection_get_output_fd( GTaskConnection *conn );

GIOChannel *
gtask_connection_get_gio_input_channel( GTaskConnection *conn );

xmlOutputBufferPtr
gtask_connection_get_xml_output_buffer( GTaskConnection *conn );

gboolean
gtask_connection_is_connected( GTaskConnection *conn );

const gchar *
gtask_connection_get_name( GTaskConnection *conn );

void
gtask_connection_set_name( GTaskConnection *conn, const gchar *name );

GTaskRole
gtask_connection_get_role( GTaskConnection *conn );

void
gtask_connection_set_role( GTaskConnection *conn, GTaskRole role );

GTaskStreamParser *
gtask_connection_get_stream_parser( GTaskConnection *conn );

GTaskEventDispatcher *
gtask_connection_get_event_dispatcher( GTaskConnection *conn );

/* @} */

#endif
