/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-error-packet.h>

/** @addtogroup GTaskErrorPacket */

/* @{ */

static gpointer parent_class;

static void
gtask_error_packet_class_init( GTaskErrorPacketClass *klass );

static void
gtask_error_packet_init( GTaskErrorPacket *packet, gpointer g_class );

static void
gtask_error_packet_finalize( GObject *obj );

static gboolean
gtask_error_packet_serialize_to( GTaskPacket *packet,
                                 GTaskConnection *conn );

GType gtask_error_packet_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskErrorPacketClass ),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_error_packet_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskErrorPacket ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_error_packet_init
        };

        type = g_type_register_static( GTASK_PACKET_TYPE,
                                       "GTaskErrorPacket",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_error_packet_class_init( GTaskErrorPacketClass *klass ) {
    GObjectClass     *gobject_class = G_OBJECT_CLASS( klass );
    GTaskPacketClass *packet_class = GTASK_PACKET_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->finalize = gtask_error_packet_finalize;

    packet_class->serialize_packet = gtask_error_packet_serialize_to;
}

static void
gtask_error_packet_init( GTaskErrorPacket *packet, gpointer g_class ) {
    GTaskErrorPacketPrivate *private = g_new0( GTaskErrorPacketPrivate, 1 );

    packet->private = private;

    private->error = GTASK_ERROR_NONE;
    private->error_msg = g_string_new( NULL );
}

static void
gtask_error_packet_finalize( GObject *obj ) {
    GTaskErrorPacket        *packet = GTASK_ERROR_PACKET( obj );
    GTaskErrorPacketPrivate *private = packet->private;

    g_string_free( private->error_msg, TRUE );

    g_free( private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

/**
 * @brief writes out the xml form of the passed in GTaskErrorPacket
 *        to the passed in GTaskConnection. In general this function
 *        should never be called, except from within the library.
 *        To serialize any packet call gtask_packet_serialize( GTaskPacket * )
 *        instead.
 *
 * @author Michael Henson
 *
 * @param packet a GTaskErrorPacket
 * @param conn the connection to serialize to
 *
 * @return the success/failure of the operation
 */
static gboolean
gtask_error_packet_serialize_to( GTaskPacket *packet, GTaskConnection *conn ) {
    GTaskErrorPacket   *e_packet;
    xmlOutputBufferPtr  xml_output;
    xmlDocPtr           doc;
    xmlNodePtr          error_node;
    xmlNodePtr          error_msg;

    g_return_val_if_fail( GTASK_IS_ERROR_PACKET( packet ), FALSE );
    g_return_val_if_fail( GTASK_IS_CONNECTION( conn ), FALSE );
    g_return_val_if_fail( gtask_connection_is_connected( conn ), FALSE );

    e_packet = (GTaskErrorPacket *) packet;

    doc = xmlNewDoc( "1.0" );

    g_return_val_if_fail( doc != NULL, FALSE );

    error_node = xmlNewNode( NULL, "error" );

    if( !error_node ) {
        xmlFreeDoc( doc );

        return FALSE;
    }

    xmlDocSetRootElement( doc, error_node );
    xmlNewProp( error_node, "type", "0" ); //packet->private->error );

    error_msg = xmlNewText( e_packet->private->error_msg->str );

    if( !error_msg ) {
        xmlFreeDoc( doc );

        return FALSE;
    }

    xmlAddChild( error_node, error_msg );

    xml_output = gtask_connection_get_xml_output_buffer( conn );
    xmlNodeDumpOutput( xml_output, doc, error_node, 1, 0, NULL );
    xmlOutputBufferFlush( xml_output );

    xmlFreeDoc( doc );

    return TRUE;
}

/**
 * @brief creates a new GTaskErrorPacket
 *
 * @author Michael Henson
 *
 * @return the new GTaskErrorPacket
 */
GTaskErrorPacket *
gtask_error_packet_new( ) {
    return g_object_new( GTASK_ERROR_PACKET_TYPE, NULL );
}

/**
 * @brief returns the error type of the passed in GTaskErrorPacket
 *
 * @author Michael Henson
 *
 * @param packet a GTaskErrorPacket
 *
 * @return the GTaskErrorType of the packet
 */
GTaskErrorType
gtask_error_packet_get_error_type( GTaskErrorPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_ERROR_PACKET( packet ), GTASK_ERROR_NONE );

    return packet->private->error;
}

/**
 * @brief sets the error type of the passed in GTaskErrorPacket 
 *
 * @author Michael Henson
 *
 * @param packet a GTaskErrorPacket
 * @param error a GTaskErrorType
 */
void
gtask_error_packet_set_error_type( GTaskErrorPacket *packet,
                                   GTaskErrorType error )
{
    g_return_if_fail( GTASK_IS_ERROR_PACKET( packet ) );

    packet->private->error = error;
}

/**
 * @brief returns a readable error message for the type of error
 *
 * @author Michael Henson
 *
 * @param packet a GTaskErrorPacket
 *
 * @return the error string
 */
const char *
gtask_error_packet_get_error_message( GTaskErrorPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_ERROR_PACKET( packet ), "" );

    return packet->private->error_msg->str;
}

/**
 * @brief sets the readable error message for the type of error
 *
 * @author Michael Henson
 *
 * @param packet a GTaskErrorPacket
 * @param error_msg the error message
 */
void
gtask_error_packet_set_error_message( GTaskErrorPacket *packet,
                                      const char *error_msg )
{
    GTaskErrorPacketPrivate *private;

    g_return_if_fail( GTASK_IS_ERROR_PACKET( packet ) );

    private = packet->private;

    /** @todo we may want to check to see if at some point and time a 
     *        stupidly large buffer was allocated and is no longer needed
     *        and free up some memory
     */
    g_string_assign( private->error_msg, error_msg );
}

/* @} */
