/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include <gtask/gtask-generic-task.h>

/** @addtogroup GTaskGenericTask */

/* @{ */

enum {
    C_PROP_0,
    C_PROP_APP,
    C_PROP_ID,
    C_PROP_TITLE,
    C_PROP_PERCENT,
    C_PROP_ACTIVITY_STATE,
    C_PROP_STATUS,
    C_PROP_ERROR,
    C_PROP_CATEGORY,
    C_PROP_THUMBNAIL
};

static gpointer parent_class;

static void
gtask_generic_task_class_init( GTaskGenericTaskClass *klass );

static void
gtask_generic_task_init( GTaskGenericTask *gtask );

static gboolean
gtask_generic_task_fill_xml_doc( GTaskGenericTask *task, xmlDocPtr doc );

static void
gtask_generic_task_dispose( GObject *obj );

static void
gtask_generic_task_finalize( GObject *obj );


static void
gtask_generic_task_set_property( GObject *object,
                                 guint param_id,
                                 const GValue *value,
                                 GParamSpec *pspec );

static void
gtask_generic_task_get_property( GObject *object,
                                 guint param_id,
                                 GValue *value,
                                 GParamSpec *pspec );

GType gtask_activity_state_get_type( ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GEnumValue values[] = {
            { GTASK_ACTIVITY_STATE_ACTIVE, "GTASK_ACTIVITY_STATE_ACTIVE", "active" },
            { GTASK_ACTIVITY_STATE_INACTIVE, "GTASK_ACTIVITY_STATE_INACTIVE", "inactive" },
            { GTASK_ACTIVITY_STATE_COMPLETED, "GTASK_ACTIVITY_STATE_COMPLETED", "completed" },
            { 0, NULL, NULL }
        };

        type = g_enum_register_static( "GTaskActivityState", values );
    }

    return type;
}

GType gtask_generic_task_get_type( ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskGenericTaskClass ),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_generic_task_class_init,  /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskGenericTask ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_generic_task_init
        };

        type = g_type_register_static( G_TYPE_OBJECT,
                                       "GTaskGenericTask",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_generic_task_init( GTaskGenericTask *gtask ) {
    GTaskGenericTaskPrivate *private;

    private = g_new0( GTaskGenericTaskPrivate, 1 );

    gtask->private = private;

    private->activity_state = GTASK_ACTIVITY_STATE_INACTIVE;

    private->application = g_string_new( NULL );
    private->task_id= g_string_new( NULL );
    private->title = g_string_new( NULL );
    private->category= g_string_new( NULL );
    private->thumbnail_uri = g_string_new( NULL );
    private->status_msg = g_string_new( NULL );
    private->error_msg = g_string_new( NULL );

    private->files = NULL;

    private->time_left = GTASK_TIME_LEFT_UNKNOWN;

    private->dispose_has_run = FALSE;
}

/**
 * @brief outputs the task to the specified connection
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param conn a GTaskConnection
 *
 * @return the success/failure of the operation
 */
gboolean
gtask_generic_task_serialize_to( GTaskGenericTask *task,
                                 GTaskConnection *conn )
{
    xmlOutputBufferPtr xml_output;
    xmlDocPtr          doc;
    xmlNodePtr         task_node;

    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), FALSE );
    g_return_val_if_fail( GTASK_IS_CONNECTION( conn ), FALSE );
    g_return_val_if_fail( !task->private->dispose_has_run, FALSE );
    g_return_val_if_fail( gtask_connection_is_connected( conn ), FALSE );

    doc = xmlNewDoc( "1.0" );

    if( !doc )
        return FALSE;

    /** @todo eventually we need to walk down the class
     * heirarchy and fill this document in a bit */
    if( !gtask_generic_task_fill_xml_doc( task, doc ) ) {
        xmlFreeDoc( doc );

        g_message( "error filling xml doc" );
        return FALSE;
    }

    task_node  = xmlDocGetRootElement( doc );
    xml_output = gtask_connection_get_xml_output_buffer( conn );

    xmlNodeDumpOutput( xml_output, doc, task_node, 1, 1, NULL );
    xmlOutputBufferFlush( xml_output );

    xmlFreeDoc( doc );

    return TRUE;
}

/**
 * @brief fills the document with at least a root node (task)
 *        and any sub elements that need to be sent to
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param doc a libxml2 document pointer
 *
 * @return the success/failure of the operation
 */
static gboolean
gtask_generic_task_fill_xml_doc( GTaskGenericTask *task, xmlDocPtr doc ) {
    GTaskGenericTaskPrivate *private = task->private;
    xmlNodePtr               task_node;
    GString                 *tmp;
    GList                   *files = private->files;

    task_node = xmlNewNode( NULL, "task" );

    if( !task_node )
        return FALSE;

    xmlDocSetRootElement( doc, task_node );

    if( !xmlNewProp( task_node, "application", private->application->str ) ) {
        g_message( "err app" );
        return FALSE;
    }

    if( !xmlNewProp( task_node, "id", private->task_id->str ) ) {
        g_message( "err id" );
        return FALSE;
    }

    while( files ) {
        GTaskFile  *file = GTASK_FILE( files->data );
        xmlNodePtr  file_node = xmlNewNode( NULL, "file" );
        char       *completed = ( gtask_file_get_completed( file ) ) ? "1" : "0";

        if( !file_node )
            return FALSE;

        if( !xmlAddChild( task_node, file_node ) )
            return FALSE;

        if( !xmlNewProp( file_node, "uri", gtask_file_get_uri( file ) ) )
            return FALSE;

        if( !xmlNewProp( file_node, "thumbnail-uri", gtask_file_get_thumbnail_uri( file ) ) )
            return FALSE;

        if( !xmlNewProp( file_node, "description", gtask_file_get_description( file ) ) )
            return FALSE;

        if( !xmlNewProp( file_node, "mime-type", gtask_file_get_mime_type( file ) ) )
            return FALSE;

        if( !xmlNewProp( file_node, "completed", completed ) )
            return FALSE;

        files = g_list_next( files );
    }

    if( !xmlNewTextChild( task_node, NULL, "title", private->title->str ) ) {
        g_message( "err title" );
        return FALSE;
    }

    if( !xmlNewTextChild( task_node, NULL, "category", private->category->str ) ) {
        g_message( "err category" );
        return FALSE;
    }

    if( !xmlNewTextChild( task_node, NULL, "thumbnail-uri", private->thumbnail_uri->str ) ) {
        g_message( "err thumn" );
        return FALSE;
    }

    tmp = g_string_new( NULL );

    g_string_printf( tmp, "%0.2f", gtask_generic_task_get_percent_done( task ) );

    if( !xmlNewTextChild( task_node, NULL, "percent-completed", tmp->str ) ) {
        g_message( "err percent" );

        g_string_free( tmp, TRUE );
        return FALSE;
    }

    if( !xmlNewTextChild( task_node, NULL, "activity-state", gtask_generic_task_get_activity_state_as_string( task ) ) ) {
        g_message( "err acti" );
        return FALSE;
    }

    if( !xmlNewTextChild( task_node, NULL, "status-message", private->status_msg->str ) ) {
        g_message( "err status" );
        return FALSE;
    }

    if( !xmlNewTextChild( task_node, NULL, "error-message", private->error_msg->str ) ) {
        g_message( "err err" );
        return FALSE;
    }

    g_string_printf( tmp, "%ld", gtask_generic_task_get_time_left( task ) );

    if( !xmlNewTextChild( task_node, NULL, "time-left", tmp->str ) ) {
        g_message( "err time left" );

        g_string_free( tmp, TRUE );
        return FALSE;
    }

    g_string_free( tmp, TRUE );

    return TRUE;
}

/**
 * @brief outputs the task packet to the default connection
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the success/failure of the operation
 */
gboolean
gtask_generic_task_serialize( GTaskGenericTask *task ) {
    GSList    *conns;
    gboolean  ret = FALSE;

    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), FALSE );
    g_return_val_if_fail( !task->private->dispose_has_run, FALSE );

    conns = gtask_connection_group_get_list( task->private->conn_group );

    while( conns ) {
        if( !gtask_generic_task_serialize_to( task, (GTaskConnection *) conns->data ) )
            ret = FALSE;

        conns = g_slist_next( conns );
    }

    return ret;
}

static void
gtask_generic_task_class_init( GTaskGenericTaskClass *klass ) {
    GObjectClass *gobject_class = G_OBJECT_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->dispose = gtask_generic_task_dispose;
    gobject_class->finalize = gtask_generic_task_finalize;

    gobject_class->get_property = gtask_generic_task_get_property;
    gobject_class->set_property = gtask_generic_task_set_property;

    g_object_class_install_property( gobject_class,
                                     C_PROP_APP,
                                     g_param_spec_string( "application",
                                                          "Application",
                                                          "The name of the application handling the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_ID,
                                     g_param_spec_string( "task-id",
                                                          "Task Id",
                                                          "The unique identifier of the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_TITLE,
                                     g_param_spec_string( "title",
                                                          "Title",
                                                          "The title of the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_THUMBNAIL,
                                     g_param_spec_string( "thumbnail-uri",
                                                          "Thumbnail Uri",
                                                          "The uri of to be used as a thumbnail (preview) for the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_PERCENT,
                                     g_param_spec_float( "percent-done",
                                                         "Percent Done",
                                                          "The percentage of the task that is completed.",
                                                          0.0, 1.0, 0.0,
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_ACTIVITY_STATE,
                                     g_param_spec_enum( "activity-state",
                                                        "Activity State",
                                                        "The activity state of the task (active, inactive, or completed).",
                                                        GTASK_ACTIVITY_STATE_TYPE,
                                                        GTASK_ACTIVITY_STATE_INACTIVE,
                                                        G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_STATUS,
                                     g_param_spec_string( "status-message",
                                                          "Status Message",
                                                          "The status message associated with the task. This field should be set to information that is deemed useful for the user.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_ERROR,
                                     g_param_spec_string( "error-message",
                                                          "Error Message",
                                                          "The error message associated with the task. This field should be set if an error has occurred and cleared once the error condition ceases to be relevant.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

    g_object_class_install_property( gobject_class,
                                     C_PROP_CATEGORY,
                                     g_param_spec_string( "category",
                                                          "Category",
                                                          "The category of the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );
}

static void
gtask_generic_task_dispose( GObject *obj ) {
    GTaskGenericTask *task = GTASK_GENERIC_TASK( obj );

    if( task->private->dispose_has_run )
        return;

    g_object_unref( task->private->conn_group );

    g_list_foreach( task->private->files, (GFunc) g_object_unref, NULL );
    g_list_free( task->private->files );

    task->private->files = NULL;

    task->private->conn_group = NULL;

    task->private->dispose_has_run = TRUE;

    if( G_OBJECT_CLASS( parent_class )->dispose )
        G_OBJECT_CLASS( parent_class )->dispose( obj );
}

static void
gtask_generic_task_finalize( GObject *obj ) {
    GTaskGenericTask        *task = GTASK_GENERIC_TASK( obj );
    GTaskGenericTaskPrivate *private = task->private;

    g_string_free( private->application, TRUE );

    g_string_free( private->task_id, TRUE );

    g_string_free( private->title, TRUE );

    g_string_free( private->thumbnail_uri, TRUE );

    g_string_free( private->status_msg, TRUE );

    g_string_free( private->error_msg, TRUE );

    g_free( task->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

static void
gtask_generic_task_set_property( GObject *object,
                                 guint param_id,
                                 const GValue *value,
                                 GParamSpec *pspec )
{
    GTaskGenericTask *task = GTASK_GENERIC_TASK( object );

    switch( param_id ) {
        case C_PROP_APP:
            gtask_generic_task_set_application( task, g_value_get_string( value ) );
            break;
        case C_PROP_ID:
            gtask_generic_task_set_id( task, g_value_get_string( value ) );
            break;
        case C_PROP_TITLE:
            gtask_generic_task_set_title( task, g_value_get_string( value ) );
            break;
        case C_PROP_PERCENT:
            gtask_generic_task_set_percent_done( task, g_value_get_float( value ) );
            break;
        case C_PROP_ACTIVITY_STATE:
            gtask_generic_task_set_activity_state( task, g_value_get_enum( value ) );
            break;
        case C_PROP_STATUS:
            gtask_generic_task_set_status_message( task, g_value_get_string( value ) );
            break;
        case C_PROP_ERROR:
            gtask_generic_task_set_error_message( task, g_value_get_string( value ) );
            break;
        case C_PROP_CATEGORY:
            gtask_generic_task_set_category( task, g_value_get_string( value ) );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}

static void
gtask_generic_task_get_property( GObject *object,
                                 guint param_id,
                                 GValue *value,
                                 GParamSpec *pspec )
{
    GTaskGenericTask        *task = GTASK_GENERIC_TASK( object );
    GTaskGenericTaskPrivate *private = task->private;

    switch( param_id ) {
        case C_PROP_APP:
            g_value_set_string( value, private->application->str );
            break;
        case C_PROP_ID:
            g_value_set_string( value, private->task_id->str );
            break;
        case C_PROP_TITLE:
            g_value_set_string( value, private->title->str );
            break;
        case C_PROP_PERCENT:
            g_value_set_float( value, private->percent_done );
            break;
        case C_PROP_ACTIVITY_STATE:
            g_value_set_enum( value, private->activity_state );
            break;
        case C_PROP_STATUS:
            g_value_set_string( value, private->status_msg->str );
            break;
        case C_PROP_ERROR:
            g_value_set_string( value, private->error_msg->str );
            break;
        case C_PROP_CATEGORY:
            g_value_set_string( value, private->category->str );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}


/**
 * @brief creates a new GTaskGenericTask
 *
 * @author Michael Henson
 *
 * @return the new GTaskGenericTask
 */
GTaskGenericTask *
gtask_generic_task_new( ) {
    return g_object_new( GTASK_GENERIC_TASK_TYPE, NULL );
}

/**
 * @brief creates a new GTaskGenericTask bound to the passed in GTaskConnection
 *
 * @attention This function will increase the reference count of the passed
 *            in GTaskConnectionGroup
 *
 * @author Michael Henson
 *
 * @param conn_group a GTaskConnectionGroup
 *
 * @return the new GTaskGenericTask
 */
GTaskGenericTask *
gtask_generic_task_new_with_connection_group( GTaskConnectionGroup *conn_group ) {
    GTaskGenericTask *task;

    g_return_val_if_fail( GTASK_IS_CONNECTION_GROUP( conn_group ), NULL );
    
    task = gtask_generic_task_new( );

    task->private->conn_group = conn_group;

    g_object_ref( conn_group );

    return task;
}

/**
 * @brief sets the name of the application that is performing the task
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param app the name of the application
 */
void
gtask_generic_task_set_application( GTaskGenericTask *task, const gchar *app ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( app != NULL );
    g_return_if_fail( !task->private->dispose_has_run );

    g_string_assign( task->private->application, app );

    g_object_notify( (GObject *) task, "application" );
}

/**
 * @brief returns the name of the application that is performing the task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the application performing the task (its name)
 */
const gchar *
gtask_generic_task_get_application( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->application->str;
}

/**
 * @brief sets the id associated with the passed in task
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param id the id of the task
 */
void
gtask_generic_task_set_id( GTaskGenericTask *task, const gchar *id ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( id != NULL );

    g_string_assign( task->private->task_id, id );

    g_object_notify( (GObject *)task, "task-id" );
}

/**
 * @brief returns the id associated with the passed in task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the id of the task
 */
const gchar *
gtask_generic_task_get_id( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->task_id->str;
}

/**
 * @brief sets the title associated with the passed in task
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param title the title of the task
 */
void
gtask_generic_task_set_title( GTaskGenericTask *task, const gchar *title ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( title != NULL );

    g_string_assign( task->private->title, title );

    g_object_notify( (GObject *) task, "title" );
}

/**
 * @brief Returns the title associated with the passed in task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the title of the task
 */
const gchar *
gtask_generic_task_get_title( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->title->str;
}

/**
 * @brief Sets the thumbnail uri of the passed in GTaskGenericTask
 *
 * @param task a GTaskGenericTask
 * @param uri the thumbnail uri
 */
void
gtask_generic_task_set_thumbnail_uri( GTaskGenericTask *task,
                                      const gchar *uri )
{
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( uri != NULL );

    g_string_assign( task->private->thumbnail_uri, uri );

    g_object_notify( (GObject *) task, "thumbnail-uri" );
}

/**
 * @brief returns the uri of the thumbnail associated with the passed in task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the uri of the thumbnail
 */
const gchar*
gtask_generic_task_get_thumbnail_uri( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->thumbnail_uri->str;
}

/**
 * @brief sets the percentage of the task that has been completed
 *
 * @attention If the passed in percentage is greater than 1.0 then the 
 *            activity state of the GTaskGenericTask will be set to
 *            GTASK_ACTIVITY_STATE_ACTIVE which in turn will cause all
 *            GTaskFiles associated with the task to have their completed
 *            property set to TRUE.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param percent the percentage of the task that has been completed
 */
void
gtask_generic_task_set_percent_done( GTaskGenericTask *task, gfloat percent ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    if( percent < 0 )
        percent = 0.0;
    else if( percent >= 1.0 ) {
        percent = 1.0;

        gtask_generic_task_set_activity_state( task,
                                               GTASK_ACTIVITY_STATE_COMPLETED );
    } else {
        gtask_generic_task_set_activity_state( task,
                                               GTASK_ACTIVITY_STATE_ACTIVE );
    }

    task->private->percent_done = percent;

    g_object_notify( (GObject *) task, "percent-done" );
}

/**
 * @brief returns the percentage of the task that has been completed
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the percent completion (between 0.0 and 1.0)
 */
gfloat
gtask_generic_task_get_percent_done( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), 0.0 );

    return task->private->percent_done;
}

/**
 * @brief sets the activity state of the passed in task
 *
 * @attention If the passed in activity state is GTASK_ACTIVITY_STATE_COMPLETED
 *            then the completed property on all GTaskFiles associated
 *            with the GTaskGenericTask will be set to TRUE.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param status the new activity status
 */
void
gtask_generic_task_set_activity_state( GTaskGenericTask *task, 
                                       GTaskActivityState state )
{
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    task->private->activity_state = state;

    if( state == GTASK_ACTIVITY_STATE_COMPLETED ) {
        GList *files = gtask_generic_task_get_files( task );

        g_warning( "=+=+=+=+=+=+=+=+=+> setting down files" );
        while( files ) {
            GTaskFile *file = (GTaskFile *) files->data;

            g_warning( "HEH" );

            gtask_file_set_completed( file, TRUE );

            files = g_list_next( files );
        }
    }

    g_object_notify( (GObject *) task, "activity-state" );
}

/**
 * @brief returns the activity state of the task
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the activity state
 */
GTaskActivityState
gtask_generic_task_get_activity_state( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ),
                          GTASK_ACTIVITY_STATE_INACTIVE );

    return task->private->activity_state;
}

/**
 * @brief returns the activity state of the task as a lowercase string
 *
 * @attention The returned string should be considered read only.
 *
 * @param task a GTaskGenericTask
 *
 * @return the activity status
 */
const gchar *
gtask_generic_task_get_activity_state_as_string( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "unknown" );

    switch( task->private->activity_state ) {
        case GTASK_ACTIVITY_STATE_ACTIVE:
            return "active";
        case GTASK_ACTIVITY_STATE_INACTIVE:
            return "inactive";
        case GTASK_ACTIVITY_STATE_COMPLETED:
            return "completed";
        default:
            break;
    }

    return "unknown";
}

/**
 * @brief attempts to set the activity state of the task from the passed in 
 *        string. If the state is unparsable, then the task is marked as
 *        GTASK_ACTIVITY_STATE_INACTIVE
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param state the activity state to set
 */
void
gtask_generic_task_set_activity_state_from_string( GTaskGenericTask *task,
                                                   const char *state )
{
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( state != NULL );

    if( strcasecmp( state, "active" ) == 0 ) {
        gtask_generic_task_set_activity_state( task,
                                               GTASK_ACTIVITY_STATE_ACTIVE );
    } else if( strcasecmp( state, "completed" ) == 0 ) {
        gtask_generic_task_set_activity_state( task,
                                               GTASK_ACTIVITY_STATE_COMPLETED );
    } else {
        gtask_generic_task_set_activity_state( task,
                                               GTASK_ACTIVITY_STATE_INACTIVE );
    }

    g_object_notify( (GObject *) task, "activity-state" );
}

/**
 * @brief sets the status message associate with the passed in task
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param msg the status message to set
 */
void
gtask_generic_task_set_status_message( GTaskGenericTask *task,
                                       const gchar *msg )
{
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( msg != NULL );

    g_string_assign( task->private->status_msg, msg );

    g_object_notify( (GObject *) task, "status-message" );
}

/**
 * @brief returns the status message associated with the passed in task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the status message or an empty string if there is none */
const gchar *
gtask_generic_task_get_status_message( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->status_msg->str;
}

/**
 * @brief sets the error message associated with the passed in task
 *
 * @author Michael Henosn
 *
 * @param task a GTaskGenericTask
 * @param msg the error message to set
 */
void
gtask_generic_task_set_error_message( GTaskGenericTask *task,
                                      const gchar *msg )
{
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( msg != NULL );

    g_string_assign( task->private->error_msg, msg );

    g_object_notify( (GObject *) task, "error-message" );
}

/**
 * @brief returns the error message associated with the passed in task
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the error message, or an empty string if there is none
 */
const gchar *
gtask_generic_task_get_error_message( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->error_msg->str;
}

/**
 * @brief sets the category of the passed in GTaskGenericTask
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 * @param cat the category of the task
 */
void
gtask_generic_task_set_category( GTaskGenericTask *task, const gchar *cat ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( cat != NULL );

    g_string_assign( task->private->category, cat );

    g_object_notify( (GObject *) task, "category" );
}

/**
 * @brief returns the category of the passed in GTaskGenericTask
 *
 * @attention The returned string should be considered read only.
 *
 * @author Michael Henson
 *
 * @param task a GTaskGenericTask
 *
 * @return the category of the task
 */
const gchar *
gtask_generic_task_get_category( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), "" );

    return task->private->category->str;
}

/**
 * @brief Prints out some information about the passed in GTaskGenericTask
 *
 * @param task a GTaskGenericTask
 */
void
gtask_generic_task_dump( GTaskGenericTask *task ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    printf( "GTaskGenericTask:\n" );
    printf( "\tApplication: %s\n", gtask_generic_task_get_application( task ) ) ;
    printf( "\tId: %s\n", gtask_generic_task_get_id( task ) ) ;
    printf( "\tCategory: %s\n", gtask_generic_task_get_category( task ) ) ;
    printf( "\tThumbnail URI: %s\n", gtask_generic_task_get_thumbnail_uri( task ) ) ;
    printf( "\tStatus Message: %s\n", gtask_generic_task_get_status_message( task ) ) ;
    printf( "\tError Message: %s\n", gtask_generic_task_get_error_message( task ) ) ;
    printf( "\tPercent Done: %f\n", gtask_generic_task_get_percent_done( task ) ) ;
    printf( "\tTime Left: %ld\n", gtask_generic_task_get_time_left( task ) ) ;
    printf( "\tActivity State: %s\n", gtask_generic_task_get_activity_state_as_string( task ) ) ;
}

/**
 * @brief Sets the amount of time left (in seconds) until the passed in
 *        GTaskGenericTask will be completed.
 *
 * @todo perhaps use GTimeVal -- do we really need a higher granularity?
 *
 * @param task a GTaskGenericTask
 * @param time_left the number of seconds remaining or GTASK_TIME_LEFT_UNKNOWN
 *                  if the amount of time remaining is unknown
 */
void
gtask_generic_task_set_time_left( GTaskGenericTask *task, glong time_left ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    task->private->time_left = ( time_left < 0 ) ? GTASK_TIME_LEFT_UNKNOWN : time_left;
}

/**
 * @brief Returns the number of seconds until the passed in GTaskGenericTask
 *        will be completed.
 *
 * @param task a GTaskGenericTask
 *
 * @return the number of seconds left
 */
glong
gtask_generic_task_get_time_left( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), GTASK_TIME_LEFT_UNKNOWN );

    return task->private->time_left;

}

/**
 * @brief Adds the passed in GTaskFile to the list of GTaskFile's associated
 *        with the passed in GTaskGenericTask.
 *
 * @attention This function will increment the reference count of the passed
 *            in GTaskFile.
 *
 * @param task a GTaskGenericTask
 * @param file a GTaskFile
 */
void
gtask_generic_task_add_file( GTaskGenericTask *task, GTaskFile *file ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );
    g_return_if_fail( GTASK_IS_FILE( file ) );

    g_object_ref( file );

    task->private->files = g_list_append( task->private->files, file );
}

/**
 * @brief Adds a the GList of GTaskFiles to the passed in GTaskGenericTask
 *
 * @attention This function will increment the reference count of the passed
 *            in GTaskFiles found in the GList.
 *
 * @param task a GTaskGenericTask
 * @param files a GList whose data elements are of type GTaskFile
 */
void
gtask_generic_task_add_files( GTaskGenericTask *task, GList *files ) {
    GList *list = files;

    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    while( list ) {
        if( !GTASK_IS_FILE( list->data ) )
            continue;

        g_object_ref( list->data );
        
        task->private->files = g_list_append( task->private->files, list->data );

        list = g_list_next( list );
    }
}

/**
 * @brief Returns a GList of GTaskFiles
 *
 * @attention The return GList should be considered read only.
 *
 * @param task a GTaskGenericTask
 *
 * @return a GList of GTaskFiles
 */
GList *
gtask_generic_task_get_files( GTaskGenericTask *task ) {
    g_return_val_if_fail( GTASK_IS_GENERIC_TASK( task ), NULL );

    return task->private->files;
}

/**
 * @brief Removes all GTaskFiles associated with the passed in GTaskGenericTask.
 *
 * @param task a GTaskGenericTask
 */
void
gtask_generic_task_clear_files( GTaskGenericTask *task ) {
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    g_list_foreach( task->private->files, (GFunc) g_object_unref, NULL );

    g_list_free( task->private->files );

    task->private->files = NULL;
}

/* @} */
