/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-packet.h>

/** @addtogroup GTaskPacket */

/* @{ */

static gpointer parent_class;

static void
gtask_packet_init( GTaskPacket *packet, gpointer g_class );

static void
gtask_packet_class_init( GTaskPacketClass *klass );

static void
gtask_packet_dispose( GObject *packet );

static void
gtask_packet_finalize( GObject *packet );

GType gtask_packet_get_type( ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskPacketClass ),
            NULL,   /* base init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_packet_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskPacket ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_packet_init
        };

        type = g_type_register_static( G_TYPE_OBJECT,
                                       "GTaskPacket",
                                        &info,
                                        0 );
    }

    return type;
}

static void
gtask_packet_init( GTaskPacket *packet, gpointer g_class ) {
    packet->private = g_new0( GTaskPacketPrivate, 1 );

    packet->private->conn_group = NULL;
}

static void
gtask_packet_class_init( GTaskPacketClass *klass ) {
    GObjectClass *gobject_class = G_OBJECT_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->dispose = gtask_packet_dispose;
    gobject_class->finalize = gtask_packet_finalize;

    klass->serialize_packet = NULL;
}

static void
gtask_packet_dispose( GObject *obj ) {
    GTaskPacket *packet = GTASK_PACKET( obj );

    if( packet->private->dispose_has_run )
        return;

   g_object_unref( packet->private->conn_group );

   packet->private->conn_group = NULL;

   packet->private->dispose_has_run = TRUE;

   if( G_OBJECT_CLASS( parent_class )->dispose )
       G_OBJECT_CLASS( parent_class )->dispose( obj );
}

static void
gtask_packet_finalize( GObject *obj ) {
    GTaskPacket *packet = GTASK_PACKET( obj );

    g_free( packet->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

/**
 * @brief prints out the packet over the default connection
 *
 * @author Michael Henson
 *
 * @param packet a GTaskPacket
 *
 * @return the success/failure of the operation (ie any connection in the group
 *         failed)
 */
gboolean
gtask_packet_serialize( GTaskPacket *packet ) {
    GSList   *conns;
    gboolean  ret = TRUE;

    g_return_val_if_fail( GTASK_IS_PACKET( packet ), FALSE );
    g_return_val_if_fail( !packet->private->dispose_has_run, FALSE );
    
    conns = gtask_connection_group_get_list( packet->private->conn_group );

    while( conns ) {
        if( !gtask_packet_serialize_to( packet, (GTaskConnection *) conns->data ) )
            ret = FALSE;

        conns = g_slist_next( conns );
    }

    return ret;
}

/**
 * @brief prints out the packet over the passed in connection
 *
 * @author Michael Henson
 *
 * @param packet a GTaskPacket
 * @param conn a GTaskConnection
 *
 * @return the success/failure of the operation
 */
gboolean
gtask_packet_serialize_to( GTaskPacket *packet, GTaskConnection *conn ) {
    g_return_val_if_fail( GTASK_IS_PACKET( packet ), FALSE );
    g_return_val_if_fail( GTASK_IS_CONNECTION( conn ), FALSE );

    if( !GTASK_PACKET_GET_CLASS(packet)->serialize_packet ) {
        g_warning( "GTASK_PACKET_GET_CLASS(packet)->serialize_packet is NULL. Perhaps this packet type was not properly subclassed?" );

        return FALSE;
    }

    return GTASK_PACKET_GET_CLASS(packet)->serialize_packet( packet, conn );
}

/**
 * @brief returns the default connection that the GTaskPacket will serialize to
 *
 * @attention This function does not add any additional references to the
 *            GTaskConnectionGroup that is returned.
 *
 * @author Michael Henson
 *
 * @param packet a GTaskPacket
 *
 * @return the GTaskConnection
 */
GTaskConnectionGroup *
gtask_packet_get_connection_group( GTaskPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_PACKET( packet ),
                          gtask_connection_group_new( ) );

    return packet->private->conn_group;
}

/**
 * @brief sets the default connection that the GTaskPacket will serialize to
 *
 * @attention This function will increase the reference count of the passed in 
 *            GTaskConnectionGroup
 *
 * @author Michael Henson
 *
 * @param packet a GTaskPacket
 * @param conn a GTaskConnection
 */
void
gtask_packet_set_connection_group( GTaskPacket *packet,
                                   GTaskConnectionGroup *conn_group ) {
    g_return_if_fail( GTASK_IS_PACKET( packet ) );
    g_return_if_fail( GTASK_IS_CONNECTION_GROUP( conn_group ) );

    if( packet->private->conn_group )
        g_object_unref( packet->private->conn_group );

    g_object_ref( conn_group );

    packet->private->conn_group = conn_group;
}

/* @} */
