/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-remove-watch-packet.h>

/** @addtogroup GTaskRemoveWatchPacket */

/* @{ */

enum {
    PROP_0,
    PROP_WATCH_ID
};

static gpointer parent_class = NULL;

static void
gtask_remove_watch_packet_class_init( GTaskRemoveWatchPacketClass *klass );

static void
gtask_remove_watch_packet_init( GTaskRemoveWatchPacket *packet,
                                gpointer g_class );

static void
gtask_remove_watch_packet_finalize( GObject *packet );

static gboolean
gtask_remove_watch_packet_serialize_to( GTaskPacket     *packet,
                                        GTaskConnection *conn );


static void
gtask_remove_watch_packet_set_property( GObject *object,
                                        guint param_id,
                                        const GValue *value,
                                        GParamSpec *pspec );

static void
gtask_remove_watch_packet_get_property( GObject *object,
                                        guint param_id,
                                        GValue *value,
                                        GParamSpec *pspec );

GType gtask_remove_watch_packet_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskRemoveWatchPacketClass ),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_remove_watch_packet_class_init, /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskRemoveWatchPacket ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_remove_watch_packet_init
        };

        type = g_type_register_static( GTASK_PACKET_TYPE,
                                       "GTaskRemoveWatchPacket",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_remove_watch_packet_class_init( GTaskRemoveWatchPacketClass *klass ) {
    GObjectClass     *gobject_class = G_OBJECT_CLASS( klass );
    GTaskPacketClass *packet_class = GTASK_PACKET_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->finalize = gtask_remove_watch_packet_finalize;
    gobject_class->set_property = gtask_remove_watch_packet_set_property;
    gobject_class->get_property = gtask_remove_watch_packet_get_property;

    packet_class->serialize_packet = gtask_remove_watch_packet_serialize_to;

    g_object_class_install_property( (GObjectClass *) klass,
                                     PROP_WATCH_ID,
                                     g_param_spec_string( "watch-id",
                                                          "Watch Id",
                                                          "The identifier of the watch that is to be removed",
                                                          "",
                                                          G_PARAM_READWRITE ) );
}

static void
gtask_remove_watch_packet_init( GTaskRemoveWatchPacket *packet,
                                gpointer g_class )
{
    packet->private = g_new0( GTaskRemoveWatchPacketPrivate, 1 );

    packet->private->watch_id = g_string_new( NULL );
}

static void
gtask_remove_watch_packet_finalize( GObject *obj ) {
    GTaskRemoveWatchPacket *packet = GTASK_REMOVE_WATCH_PACKET( obj );

    g_string_free( packet->private->watch_id, TRUE );

    g_free( packet->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

static void
gtask_remove_watch_packet_set_property( GObject *object,
                                        guint param_id,
                                        const GValue *value,
                                        GParamSpec *pspec  )
{
    GTaskRemoveWatchPacket *packet = GTASK_REMOVE_WATCH_PACKET( object );

    switch( param_id ) {
        case PROP_WATCH_ID:
            gtask_remove_watch_packet_set_watch_id( packet,
                                                  g_value_get_string( value ) );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}

static void
gtask_remove_watch_packet_get_property( GObject *object,
                                        guint param_id,
                                        GValue *value,
                                        GParamSpec *pspec  )
{
    GTaskRemoveWatchPacket *packet = GTASK_REMOVE_WATCH_PACKET( object );

    switch( param_id ) {
        case PROP_WATCH_ID:
            g_value_set_string( value, packet->private->watch_id->str );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}

/** 
 * @brief writes out the xml form of the passed in GTaskRemoveWatchPacket
 *        to the passed in GTaskConnection. In general this function
 *        should never be called, except from within the library.
 *        To serialize any packet call gtask_packet_serialize( GTaskPacket * )
 *        instead.
 *
 * @author Michael Henson
 *
 * @param packet a GTaskRemoveWatchPacket
 * @param conn a GTaskConnection
 *
 * @return the success/failure of the operation
 */
static gboolean
gtask_remove_watch_packet_serialize_to( GTaskPacket     *packet,
                                        GTaskConnection *conn )
{
    GTaskRemoveWatchPacket *g_packet;
    xmlOutputBufferPtr      xml_output;
    xmlDocPtr               doc;
    xmlNodePtr              node;

    g_return_val_if_fail( GTASK_IS_REMOVE_WATCH_PACKET( packet ), FALSE );
    g_return_val_if_fail( GTASK_IS_CONNECTION( conn ), FALSE );
    g_return_val_if_fail( gtask_connection_is_connected( conn ), FALSE );

    g_packet = ( GTaskRemoveWatchPacket * ) packet;

    doc = xmlNewDoc( "1.0" );

    if( !doc )
        return FALSE;

    node = xmlNewNode( NULL, "remove-watch" );

    if( !node ) {
        xmlFreeDoc( doc );

        return FALSE;
    }

    xmlDocSetRootElement( doc, node );
    if( !xmlNewProp( node, "id", g_packet->private->watch_id->str ) ) {
        xmlFreeDoc( doc );

        return FALSE;
    }

    xml_output = gtask_connection_get_xml_output_buffer( conn );
    xmlNodeDumpOutput( xml_output, doc, node, 1, 0, NULL );
    xmlOutputBufferFlush( xml_output );

    xmlFreeDoc( doc );

    return TRUE;
}

/**
 * @brief creates a new GTaskRemoveWatchPacket
 *
 * @author Michael Henson
 *
 * @return the new GTaskRemoveWatchPacket
 */
GTaskRemoveWatchPacket *
gtask_remove_watch_packet_new( ) {
    return g_object_new( GTASK_REMOVE_WATCH_PACKET_TYPE, NULL );
}

/**
 * @brief creates a new GTaskRemoveWatchPacket with the passed in watch_id
 *
 * @author Michael Henson
 *
 * @param watch_id the watch id
 *
 * @return the new GTaskRemoveTaskPacket
 */
GTaskRemoveWatchPacket *
gtask_remove_watch_packet_new_with_watch_id( const char *watch_id ) {
    GTaskRemoveWatchPacket *packet = gtask_remove_watch_packet_new( );

    gtask_remove_watch_packet_set_watch_id( packet, watch_id );

    return packet;
}

/**
 * @brief returns the watch id held by the passed in GTaskRemoveWatchPacket
 *
 * @author Michael Henson
 *
 * @param packet a GTaskRemoveWatchPacket
 *
 * return the watch id
 */
const char *
gtask_remove_watch_packet_get_watch_id( GTaskRemoveWatchPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_REMOVE_WATCH_PACKET( packet ), "" );

    return packet->private->watch_id->str;
}

/**
 * @brief sets the watch id of the passed in GTaskRemoveWatchPacket
 *
 * @author Michael Henson
 *
 * @param packet a GTaskRemoveWatchPacket
 * @param watch_id the new watch id
 */
void
gtask_remove_watch_packet_set_watch_id( GTaskRemoveWatchPacket *packet,
                                         const char *watch_id )
{
    g_return_if_fail( GTASK_IS_REMOVE_WATCH_PACKET( packet ) );
    g_return_if_fail( watch_id != NULL );

    g_string_assign( packet->private->watch_id, watch_id );

    g_object_notify( (GObject *) packet, "watch-id" );
}

/* @} */
