/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-task-packet.h>

/** @addtogroup GTaskTaskPacket */

/* @{ */

static gpointer parent_class;

static void
gtask_task_packet_class_init( GTaskTaskPacketClass *klass );

static void
gtask_task_packet_init( GTaskTaskPacket *packet, gpointer g_class );

static void
gtask_task_packet_dispose( GObject *obj );

static void
gtask_task_packet_finalize( GObject *obj );

static gboolean
gtask_task_packet_serialize_to( GTaskPacket     *packet,
                                GTaskConnection *conn );

GType gtask_task_packet_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskTaskPacketClass ),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_task_packet_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskTaskPacket ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_task_packet_init
        };

        type = g_type_register_static( GTASK_PACKET_TYPE,
                                       "GTaskTaskPacket",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_task_packet_class_init( GTaskTaskPacketClass *klass ) {
    GObjectClass     *gobject_class = G_OBJECT_CLASS( klass );
    GTaskPacketClass *packet_class = GTASK_PACKET_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->dispose = gtask_task_packet_dispose;
    gobject_class->finalize = gtask_task_packet_finalize;

    packet_class->serialize_packet = gtask_task_packet_serialize_to;
}

static void
gtask_task_packet_init( GTaskTaskPacket *packet, gpointer g_class ) {
    GTaskTaskPacketPrivate *private = g_new0( GTaskTaskPacketPrivate, 1 );

    packet->private = private;
    private->task = NULL;
    private->dispose_has_run = FALSE;
}

/** @todo do we have to chain up to our parent's dispose method too? */
static void
gtask_task_packet_dispose( GObject *obj ) {
    GTaskTaskPacket        *packet = GTASK_TASK_PACKET( obj );
    GTaskTaskPacketPrivate *private = packet->private;

    if( packet->private->dispose_has_run )
        return;

    if( private->task ) {
        g_object_unref( private->task );

        private->task = NULL;
    }

    packet->private->dispose_has_run = TRUE;

    if( G_OBJECT_CLASS( parent_class )->dispose )
        G_OBJECT_CLASS( parent_class )->dispose( obj );
}

static void
gtask_task_packet_finalize( GObject *obj ) {
    GTaskTaskPacket *packet = GTASK_TASK_PACKET( obj );

    g_free( packet->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

/**
 * @brief writes out the xml form of the passed in GTaskTaskPacket to
 *        the passed in GTaskConnection. In general this function
 *        should never be called, except from within the library.
 *        To serialize any packet call gtask_packet_serialize( GTaskPacket * )
 *        instead.
 *
 * @author Michael Henson
 *  
 * @param packet a GTaskTaskPacket
 * @param conn the connection to serialize to
 *
 * @return the success/failure of the operation
 */
static gboolean
gtask_task_packet_serialize_to( GTaskPacket     *packet,
                                GTaskConnection *conn )
{
    g_return_val_if_fail( GTASK_IS_TASK_PACKET( packet ), FALSE );

    return gtask_generic_task_serialize_to( ((GTaskTaskPacket *) packet)->private->task, conn );
}

/**
 * @brief creates a new GTaskTaskPacket that holds
 *        the passed in task
 *
 * @attention This function will increase the reference count of the passed in
 *            GTaskGenericTask.
 *
 * @param task a GTaskGenericTask
 *
 * @return the new GTaskTaskPacket
 */
GTaskTaskPacket *
gtask_task_packet_new_with_task( GTaskGenericTask *task ) {
    GTaskTaskPacket *packet;

    packet = g_object_new( GTASK_TASK_PACKET_TYPE, NULL );

    packet->private->task = task;

    g_object_ref( task );

    return packet;
}

/**
 * @brief returns the task held by the packet
 *
 * @attention This function will not increase the reference count of the 
 *            GTaskGenericTask that is returned.
 *
 * @author Michael Henson
 *
 * @param packet a GTaskTaskPacket
 *
 * @return the GTaskGenericTask
 */
GTaskGenericTask *
gtask_task_packet_get_task( GTaskTaskPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_TASK_PACKET( packet ), NULL );
    g_return_val_if_fail( !packet->private->dispose_has_run, NULL );

    return packet->private->task;
}

/* @} */
