/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask/gtask-watch-packet.h>

/** @addtogroup GTaskWatchPacket */

/* @{ */

gpointer parent_class;

static void
gtask_watch_packet_class_init( GTaskWatchPacketClass *klass );

static void
gtask_watch_packet_init( GTaskWatchPacket *packet, gpointer g_class );

static void
gtask_watch_packet_finalize( GObject *packet );

static gboolean
gtask_watch_packet_serialize_to( GTaskPacket     *packet,
                                 GTaskConnection *conn );


GType gtask_watch_packet_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskWatchPacketClass ),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_watch_packet_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskWatchPacket ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_watch_packet_init
        };

        type = g_type_register_static( GTASK_PACKET_TYPE,
                                       "GTaskWatchPacket",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_watch_packet_class_init( GTaskWatchPacketClass *klass ) {
    GObjectClass     *gobject_class = G_OBJECT_CLASS( klass );
    GTaskPacketClass *packet_class = GTASK_PACKET_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->finalize = gtask_watch_packet_finalize;

    packet_class->serialize_packet = gtask_watch_packet_serialize_to;
}

static void
gtask_watch_packet_init( GTaskWatchPacket *packet, gpointer g_class ) {
    packet->private = g_new0( GTaskWatchPacketPrivate, 1 );

    packet->private->watch_id = g_string_new( NULL );
}

static void
gtask_watch_packet_finalize( GObject *obj ) {
    GTaskWatchPacket *packet = GTASK_WATCH_PACKET( obj );

    g_string_free( packet->private->watch_id, TRUE );

    g_free( packet->private );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

/**
 * @brief writes out the xml form of the passed in GTaskWatchPacket to
 *        the passed in GTaskConnection. In general this function
 *        should never be called, except from within the library.
 *        To serialize any packet call gtask_packet_serialize( GTaskPacket * )
 *        instead.
 *
 * @author Michael Henson
 *  
 * @param packet a GTaskRolePacket
 * @param conn the connection to serialize to
 *
 * @return the success/failure of the operation
 */
static gboolean
gtask_watch_packet_serialize_to( GTaskPacket *packet,
                                 GTaskConnection *conn ) 
{
    GTaskWatchPacket   *w_packet;
    xmlOutputBufferPtr  xml_output;
    xmlDocPtr           doc;
    xmlNodePtr          watch_node;

    g_return_val_if_fail( GTASK_IS_WATCH_PACKET( packet ), FALSE );
    g_return_val_if_fail( GTASK_IS_CONNECTION( conn ), FALSE );
    g_return_val_if_fail( gtask_connection_is_connected( conn ), FALSE );

    w_packet = (GTaskWatchPacket *) packet;

    doc = xmlNewDoc( "1.0" );

    if( !doc )
        return FALSE;

    watch_node = xmlNewNode( NULL, "watch" );

    if( !watch_node ) {
        xmlFreeDoc( doc );

        return FALSE;
    }

    xmlDocSetRootElement( doc, watch_node );

    if( !xmlNewProp( watch_node, "id", w_packet->private->watch_id->str ) ) {
        xmlFreeDoc( doc );

        return FALSE;
    }


    xml_output = gtask_connection_get_xml_output_buffer( conn );
    xmlNodeDumpOutput( xml_output, doc, watch_node, 1, 0, NULL );
    xmlOutputBufferFlush( xml_output );

    xmlFreeDoc( doc );

    return TRUE;
}

/**
 * @brief creates a new GTaskWatchPacket
 *
 * @author Michael Henson
 *
 * @return the new GTaskWatchPacket
 */
GTaskWatchPacket *
gtask_watch_packet_new( ) {
    return g_object_new( GTASK_WATCH_PACKET_TYPE, NULL );
}

/**
 * @brief creates a new GTaskWatchPacket with the passed in watch_id
 *
 * @author Michael Henson
 *
 * @param watch_id the watch id
 *
 * @return the new GTaskWatchPacket
 */
GTaskWatchPacket *
gtask_watch_packet_new_with_watch_id( const char *watch_id ) {
    GTaskWatchPacket *packet = gtask_watch_packet_new( );

    gtask_watch_packet_set_watch_id( packet, watch_id );

    return packet;
}

/**
 * @brief returns the watch id held by the passed in GTaskWatchPacket
 *
 * @author Michael Henson
 *
 * @param packet a GTaskWatchPacket
 *
 * @return the watch id
 */
const char *
gtask_watch_packet_get_watch_id( GTaskWatchPacket *packet ) {
    g_return_val_if_fail( GTASK_IS_WATCH_PACKET( packet ), "" );

    return packet->private->watch_id->str;
}

/**
 * @brief sets the watch id of the passed GTaskWatchPacket
 *
 * @author Michael Henson
 *
 * @param packet a GTaskWatchPacket
 * @param watch_id the new watch id
 */
void
gtask_watch_packet_set_watch_id( GTaskWatchPacket *packet,                                                       const char *watch_id )
{
    g_return_if_fail( GTASK_IS_WATCH_PACKET( packet ) );

    g_string_assign( packet->private->watch_id, watch_id );

    /** @todo implement me */
//    g_object_notify( (GObject *) packet, "watch-id" );
}

/* @} */
